import datetime
from urllib.parse import urlparse
from urllib.parse import parse_qs

from .errors import BRPaginationError
from .utils import StackableFinder, Localizer


def _get_object(lst, _id):
    """
    Internal function to grab data referenced inside response['included']
    """
    for item in lst:
        if item['id'] == _id:
            return item


class BaseBRObject:
    """
    A base object for most data classes
    
    Attributes
    ----------
    id : int
        A general unique ID for each type of data.
    """
    __slots__ = ['id']

    def __init__(self, data):
        self.id = data['id']


class Player(BaseBRObject):
    """
    A class that holds general user data, if this is through a Match, this will not have
    name, picture and title, only an ID
    
    Attributes
    ----------
    id : int
        A general unique ID for each type of data.
    name : str
        The player's name
    picture : int
        The picture ID for this player
    title : int
        This player's ingame title
    """
    __slots__ = ['id', 'name', 'picture', 'title', 'stats']

    def __init__(self, data, lang: str='English'):
        super().__init__(data)
        if data.get('attributes'):
            self.name = data['attributes']['name']
            self.picture = data['attributes']['stats'].pop('picture')
            self.title = data['attributes']['stats'].pop('title')
            stackables = StackableFinder()
            localizer = Localizer(lang)
            self.stats = {}
            for key, value in data['attributes']['stats'].items():
                _item = stackables.find(key)
                if _item is not None:
                    name = localizer.localize(_item['LocalizedName']) if _item['LocalizedName'] else _item['DevName']
                    self.stats[key] = {'localized_name': name,
                                       'xp': value}
                    if _item['LocalizedName']:
                        self.stats[key]['loc_id'] = _item['LocalizedName']

    def __repr__(self):
        return "<Player: id={}>".format(self.id)


class Team(BaseBRObject):
    __slots__ = ['name', 'shard_id']

    def __init__(self, data):
        super().__init__(data)
        self.name = data['name']
        self.shard_id = data['shardId']


class Participant(BaseBRObject):
    """
    A class that holds data about a participant in a match.
        
    Attributes
    ----------
    id : int
        A general unique ID for each type of data.
    actor : int
    shard_id : str
    attachment : int
    emote : int
    mount : int
    outfit : int
    side : int
    ability_uses : Optional[int]
    damage_done : Optional[int]
    damage_received : Optional[int]
    deaths : Optional[int]
    disables_done : Optional[int]
    disables_received : Optional[int]
    energy_gained : Optional[int]
    energy_used : Optional[int]
    healing_done : Optional[int]
    healing_received : Optional[int]
    kills : Optional[int]
    score : Optional[int]
    time_alive : Optional[int]
    user_id : Optional[int]
    player : Optional[:class:`Player`]
    """
    __slots__ = ['actor', 'shard_id', 'attachment', 'emote', 'mount', 'outfit', 'player', 'side', 'ability_uses',
                 'damage_done', 'damage_received', 'deaths', 'disables_done', 'disables_received', 'energy_gained',
                 'energy_used', 'healing_done', 'healing_received', 'kills', 'score', 'time_alive', 'user_id']

    def __init__(self, participant, included):
        super().__init__(participant)
        data = _get_object(included, participant['id'])
        self.actor = data['attributes']['actor']
        self.shard_id = data['attributes']['shardId']
        stats = data['attributes']['stats']
        self.attachment = stats['attachment']
        self.emote = stats['emote']
        self.mount = stats['mount']
        self.outfit = stats['outfit']
        self.side = stats['side']

        # These may or may not exist
        self.ability_uses = stats.get('abilityUses')
        self.damage_done = stats.get('damageDone')
        self.damage_received = stats.get('damageReceived')
        self.deaths = stats.get('deaths')
        self.disables_done = stats.get('disablesDone')
        self.disables_received = stats.get('disablesReceived')
        self.energy_gained = stats.get('energyGained')
        self.energy_used = stats.get('energyUsed')
        self.healing_done = stats.get('healingDone')
        self.healing_received = stats.get('healingReceived')
        self.kills = stats.get('kills')
        self.score = stats.get('score')
        self.time_alive = stats.get('timeAlive')
        self.user_id = stats.get('userID')

        if 'relationships' in data:
            self.player = Player(data['relationships']['player']['data'])
        else:
            self.player = None

    def __repr__(self):
        return "<Participant: id={0} shard_id={1} actor={2} bot={3}>".format(self.id, self.shard_id, self.actor,
                                                                             True if not self.player else False)


class Round(BaseBRObject):
    """
    A class that holds general data about a round.

    Attributes
    ----------
    id : int
        A general unique ID for each type of data.
    duration : int
        Duration in seconds of this round.
    ordinal : int
        The round number this round was in a match.
    winning_team : int
        Signifies which of the teams/rosters won the round.
    """
    __slots__ = ['duration', 'ordinal', 'winning_team', 'participants']

    def __init__(self, _round, included):
        super().__init__(_round)
        data = _get_object(included, _round['id'])
        self.duration = data['attributes']['duration']
        self.ordinal = data['attributes']['ordinal']
        self.winning_team = data['attributes']['stats']['winningTeam']

    def __repr__(self):
        return "<Round: id={} duration={} ordinal={}>".format(self.id, self.duration, self.ordinal)


class Roster(BaseBRObject):
    """
    A class that holds data about one of the two teams in a match.

    Attributes
    ----------
    id : int
        A general unique ID for each type of data.
    shard_id : str
        The shard on which this roster data resides, only `global` is available at the moment.
    score : int
        The roster's overall score for a match.
    won : bool
        Signifies if the roster won a match.
    participants : list
       A list of :class:`Participant` representing players that are part of the roster.
    """
    __slots__ = ['shard_id', 'score', 'won', 'participants', 'team']

    def __init__(self, roster, included):
        super().__init__(roster)
        data = _get_object(included, roster['id'])
        self.shard_id = data['attributes']['shardId']
        self.score = data['attributes']['stats']['score']
        self.won = True if data['attributes']['won'] == 'true' else False

        self.participants = []
        for participant in data['relationships']['participants']['data']:
            self.participants.append(Participant(participant, included))

    def __repr__(self):
        return "<Roster: id={0} shard_id={1} won={2}>".format(self.id, self.shard_id, self.won)


class MatchBase(BaseBRObject):
    """
    A class that holds data for a match.

    .. _datetime.datetime: https://docs.python.org/3.6/library/datetime.html#datetime-objects
    .. _aiohttp.ClientSession: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

    Attributes
    ----------
    id : int
        A general unique ID for each type of data.
    created_at : datetime.datetime_
        Time when the match was created.
    duration : int
        The match's duration in seconds.
    game_mode : str
        The gamemode this match was of.
    patch : str
        The Battlerite patch version this match was played on.
    shard_id : str
        The shard on which this match data resides, only `global` is available at the moment.
    map_id : int
        The match's map's ID.
    type : str
        The match type, ex: `QUICK2V2 `.
    rosters : list
        A list of :class:`Roster` objects representing rosters taking part in the match.
    rounds : list
        A list of :class:`Round` objects representing each round in the match.
    spectators : list
        A list of :class:`Participant` objects representing spectators, this seems unimplemented in the game as of now.
    telemetry_url : str
        URL for the telemetry file for this match
    session : Optional[aiohttp.ClientSession_]
        Optional session to use to request match data.
    """
    __slots__ = ['created_at', 'duration', 'game_mode', 'patch', 'shard_id', 'map_id', 'type', 'telemetry_url',
                 'rosters', 'rounds', 'spectators', 'session']

    def __init__(self, data, session, included=None):
        included = included or data['included']
        data = data if included else data['data']
        super().__init__(data)
        self.created_at = datetime.datetime.strptime(data['attributes']['createdAt'], "%Y-%m-%dT%H:%M:%SZ")
        self.duration = data['attributes']['duration']
        self.game_mode = data['attributes']['gameMode']
        self.patch = data['attributes']['patchVersion']
        self.shard_id = data['attributes']['shardId']
        self.map_id = data['attributes']['stats']['mapID']
        self.type = data['attributes']['stats']['type']
        self.rosters = []
        for roster in data['relationships']['rosters']['data']:
            self.rosters.append(Roster(roster, included))
        self.rounds = []
        for _round in data['relationships']['rounds']['data']:
            self.rounds.append(Round(_round, included))
        self.spectators = []
        for participant in data['relationships']['spectators']['data']:
            self.spectators.append(Participant(participant, included))
        self.telemetry_url = _get_object(included,
                                         data['relationships']['assets']['data'][0]['id'])['attributes']['URL']
        self.session = session


class AsyncMatch(MatchBase):
    """
    Extends :class:`MatchBase` to add async :meth:`get_telemetry`.
    """
    def __init__(self, data, session, included=None):
        super().__init__(data, session, included)

    def __repr__(self):
        return "<AsyncMatch: id={0} shard_id={1}>".format(self.id, self.shard_id)

    async def get_telemetry(self, session=None):
        """
        Get telemetry data for a match.

        .. _aiohttp.ClientSession: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

        Parameters
        ----------
        session : Optional[aiohttp.ClientSession_]
            Optional session to use to request telemetry data.

        Returns
        -------
        `dict`
            Match telemetry data
        """
        sess = session or self.session
        async with sess.get(self.telemetry_url, headers={'Accept': 'application/json'}) as resp:
            data = await resp.json()

        # After understanding the telemetry structure, to provide it as usable data is going to be a tough ordeal,
        # but one that can be looked into later
        return data


class Match(MatchBase):
    """
    Extends :class:`MatchBase` to add :meth:`get_telemetry`
    """
    def __init__(self, data, session, included=None):
        super().__init__(data, session, included)

    def __repr__(self):
        return "<Match: id={0} shard_id={1}>".format(self.id, self.shard_id)

    def get_telemetry(self, session=None):
        """
        Get telemetry data for a match.

        .. _requests.Session: http://docs.python-requests.org/en/master/api/#request-sessions

        Parameters
        ----------
        session : Optional[requests.Session_]
            Optional session to use to request telemetry data.

        Returns
        -------
        `dict`
            Match telemetry data
        """
        sess = session or self.session
        with sess.get(self.telemetry_url, headers={'Accept': 'application/json'}) as resp:
            data = resp.json()

        # After understanding the telemetry structure, to provide it as usable data is going to be a tough ordeal,
        # but one that can be looked into later
        return data


class Paginator:
    """
    Returned only by BRClient.get_matches
    """
    __slots__ = ['matches', 'next_url', 'first_url', 'client', 'prev_url', 'offset']

    def __init__(self, matches, data, client):
        self.matches = matches
        self.next_url = data.get('next')
        self.first_url = data.get('first')
        self_params = parse_qs(urlparse(data['self'])[4])
        self.offset = self_params.get('page[offset]', 0)
        if self.offset:
            self.offset = self.offset[0]
        self.prev_url = data.get('prev')
        self.client = client

    def __getitem__(self, item):
        return self.matches[item]

    def __iter__(self):
        return iter(self.matches)


class AsyncMatchPaginator(Paginator):
    """
    Returned only by the *get_matches* method of the async client.
    """
    def __init__(self, matches, data, client):
        super().__init__(matches, data, client)

    def __repr__(self):
        return "<AsyncMatchPaginator: offset={0} next={1} prev={2}>".format(self.offset,
                                                                            True if self.next_url else False,
                                                                            True if self.prev_url else False)

    async def _matchmaker(self, url, sess=None):
        data = await self.client.gen_req("{}matches".format(url), session=sess)
        matches = []
        for match in data['data']:
            matches.append(AsyncMatch(match, self.client.session, data['included']))
        print(matches)
        self.__init__(matches, data['links'], self.client)
        return matches

    async def next(self, session=None):
        """
        Move to the next page of matches.

        .. _aiohttp.ClientSession: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

        Parameters
        ----------
        session : Optional[aiohttp.ClientSession_]
            Optional session to use to make this request.

        Returns
        -------
        `list`
            A list of :class:`Match`.

        Raises
        ------
        BRPaginationError
            The current page is the last page of results
        """
        if self.next_url:
            matches = await self._matchmaker(self.next_url, session)
            return matches
        else:
            raise BRPaginationError("This is the last page")

    async def first(self, session=None):
        """
        Move to the first page of matches.

        .. _aiohttp.ClientSession: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

        Parameters
        ----------
        session : Optional[aiohttp.ClientSession_]
            Optional session to use to make this request.

        Returns
        -------
        `list`
            A list of :class:`Match`.

        """
        matches = await self._matchmaker(self.first_url, session)
        return matches

    async def prev(self, session=None):
        """
        Move to the previous page of matches.

        .. _aiohttp.ClientSession: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

        Parameters
        ----------
        session : Optional[aiohttp.ClientSession_]
            Optional session to use to make this request.

        Returns
        -------
        `list`
            A list of :class:`Match`.

        Raises
        ------
        BRPaginationError
            The current page is the first page of results
        """
        if self.prev_url:
            matches = await self._matchmaker(self.prev_url, session)
            return matches
        else:
            raise BRPaginationError("This is the first page")


class MatchPaginator(Paginator):
    """
    Returned only by the *get_matches* method of the client.
    """
    def __init__(self, matches, data, client):
        super().__init__(matches, data, client)

    def __repr__(self):
        return "<MatchPaginator: offset={0} next={1} prev={2}>".format(self.offset, True if self.next_url else False,
                                                                       True if self.prev_url else False)

    def _matchmaker(self, url, sess=None):
        data = self.client.gen_req("{}matches".format(url), session=sess)
        matches = []
        for match in data['data']:
            matches.append(Match(match, self.client.session, data['included']))
        print(matches)
        self.__init__(matches, data['links'], self.client)
        return matches

    def next(self, session=None):
        """
        Move to the next page of matches.

        .. _requests.Session: http://docs.python-requests.org/en/master/api/#request-sessions

        Parameters
        ----------
        session : Optional[requests.Session_]
            Optional session to use to make this request.

        Returns
        -------
        `list`
            A list of :class:`Match`.

        Raises
        ------
        BRPaginationError
            The current page is the last page of results
        """
        if self.next_url:
            matches = self._matchmaker(self.next_url, session)
            return matches
        else:
            raise BRPaginationError("This is the last page")

    def first(self, session=None):
        """
        Move to the first page of matches.

        .. _requests.Session: http://docs.python-requests.org/en/master/api/#request-sessions

        Parameters
        ----------
        session : Optional[requests.Session_]
            Optional session to use to make this request.

        Returns
        -------
        `list`
            A list of :class:`Match`.

        """
        matches = self._matchmaker(self.first_url, session)
        return matches

    def prev(self, session=None):
        """
        Move to the previous page of matches.

        .. _requests.Session: http://docs.python-requests.org/en/master/api/#request-sessions

        Parameters
        ----------
        session : Optional[requests.Session_]
            Optional session to use to make this request.

        Returns
        -------
        `list`
            A list of :class:`Match`.

        Raises
        ------
        BRPaginationError
            The current page is the first page of results
        """
        if self.prev_url:
            matches = self._matchmaker(self.prev_url, session)
            return matches
        else:
            raise BRPaginationError("This is the first page")



import datetime

from .errors import BRFilterException


class ClientBase:
    avl_langs = ['Brazilian', 'English', 'French', 'German', 'Italian',
                 'Japanese', 'Korean', 'Polish', 'Romanian', 'Russian',
                 'SChinese', 'Spanish', 'Turkish']

    @staticmethod
    def _isocheck(time):
        """
        Check if a time string is compatible with iso8601
        """
        try:
            datetime.datetime.strptime(time, "%Y-%m-%dT%H:%M:%SZ")
            return True
        except ValueError:
            return False

    def prepare_match_params(self, offset, limit, after, before, playerids):
        if all((after, before)):
            if all((isinstance(after, datetime.datetime), isinstance(before, datetime.datetime))):
                if before <= after:
                    raise BRFilterException("'after' must occur at a time before 'before'")
                else:
                    after = after.strftime("%Y-%m-%dT%H:%M:%SZ")
                    before = before.strftime("%Y-%m-%dT%H:%M:%SZ")
            elif all((isinstance(after, str), isinstance(before, str))):
                if not all(map(self._isocheck, (after, before))):
                    raise BRFilterException("'after' or 'before', if instances of 'str', should follow the "
                                            "'iso8601' format '%Y-%m-%dT%H:%M:%SZ'")
                t_after = datetime.datetime.strptime(after, "%Y-%m-%dT%H:%M:%SZ")
                t_before = datetime.datetime.strptime(before, "%Y-%m-%dT%H:%M:%SZ")
                if t_before <= t_after:
                    raise BRFilterException("'after' must occur at a time before 'before'")
            else:
                raise BRFilterException("'after' and 'before' should both be instances "
                                        "of either 'str' or 'datetime.datetime'")
        else:
            if after:
                if isinstance(after, datetime.datetime):
                    after = after.strftime("%Y-%m-%dT%H:%M:%SZ")
                elif not self._isocheck(after):
                    raise BRFilterException("'after', if instance of 'str', should follow the 'iso8601' format "
                                            "'%Y-%m-%dT%H:%M:%SZ'")
            elif before:
                if isinstance(before, datetime.datetime):
                    before = before.strftime("%Y-%m-%dT%H:%M:%SZ")
                elif not self._isocheck(before):
                    raise BRFilterException("'before', if instance of 'str', should follow the 'iso8601' format "
                                            "'%Y-%m-%dT%H:%M:%SZ'")
        # # Make sure 'playernames' is a list of 'str's
        # if playernames:
        #     if not all([isinstance(player, str) for player in playernames]):
        #         raise BRFilterException("'playernames' must be a list of 'str's")
        # # Make sure 'teamnames' is a list of 'str's
        # if teamnames:
        #     if not all([isinstance(team, str) for team in teamnames]):
        #         raise BRFilterException("'teamnames' must be a list of 'str's")
        params = {}
        if offset:
            params['page[offset]'] = offset
        if limit:
            params['page[limit]'] = limit
        if after:
            params['filter[createdAt-start]'] = after
        if before:
            params['filter[createdAt-end]'] = before
        # if playernames:
        #     params['filter[playerNames]'] = ','.join(playernames)
        if playerids:
            params['filter[playerIds]'] = ','.join([str(_id) for _id in playerids])
        # if teamnames:
        #     params['filter[teamNames]'] = ','.join(teamnames)

        return params

    @staticmethod
    def prepare_players_params(playerids, steamids, usernames):
        if not any((playerids, steamids, usernames)):
            raise BRFilterException("One of the filters 'playerids', 'steamids' and 'usernames' is required.")
        try:
            if len(playerids) > 6:
                raise BRFilterException("Only a maximum of 6 playerIDs are allowed for a single"
                                        " request of get_players.")
        except TypeError:
            pass
        try:
            if len(steamids) > 6:
                raise BRFilterException("Only a maximum of 6 steamIDs are allowed for a single"
                                        " request of get_players.")
        except TypeError:
            pass
        try:
            if len(usernames) > 6:
                raise BRFilterException("Only a maximum of 6 steamIDs are allowed for a single"
                                        " request of get_players.")
        except TypeError:
            pass
        params = {}
        if playerids:
            params['filter[playerIds]'] = ','.join([str(_id) for _id in playerids])
        if steamids:
            params['filter[steamIds]'] = ','.join([str(_id) for _id in steamids])
        if usernames:
            params['filter[playerNames]'] = ','.join([str(name) for name in usernames])

        return params


import json
import os

from configparser import ConfigParser


class StackableFinder:
    def __init__(self):
        _dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', 'stackables.json')
        with open(_dir) as f:
            self.data = json.load(f)

    def find(self, _id):
        for item in self.data["Mappings"]:
            if item["StackableId"] == int(_id):
                return item
        return None


class LocParser(ConfigParser):

    def to_dict(self):
        d = dict(self._sections)
        for k in d:
            d[k] = dict(self._defaults, **d[k])
            d[k].pop('__name__', None)
        return d


class Localizer:
    """
    Use this to manually localize any data with an available `loc_id`.
    
    Parameters
    ----------
    lang : str
        The language to localise game specific strings in.\n
        Currently available languages are:\n
        `Brazilian, English, French, German, Italian, Japanese, Korean,
        Polish, Romanian, Russian, SChinese, Spanish, Turkish.`
    """
    def __init__(self, lang):
        parser = LocParser()
        _dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data', 'localization',
                            '{}.ini'.format(lang))
        with open(_dir) as fp:
            parser.readfp(fp)
        self.data = parser.to_dict()['Loc']

    def localize(self, _id):
        """
        Call to return a localized string for your id.
        
        Parameters
        ----------
        _id : str
            An id that refers to the name of an object in multiple languages.\n
            Example: '035ad4c27697469e8163040ae0a4f796'
        
        Returns
        -------
        str
            A localized string in the :class:`Localizer`'s set language.
        """
        return self.data[_id]




import asyncio
import aiohttp

from .clientbase import ClientBase
from .models import Player, AsyncMatch, AsyncMatchPaginator
from .errors import BRRequestException
from .errors import NotFoundException
from .errors import BRServerException
from .errors import EmptyResponseException


class AsyncClient(ClientBase):
    """
    Top level class for user to interact with the API.

    .. _aiohttp.ClientSession: https://aiohttp.readthedocs.io/en/stable/client_reference.html#client-session

    Parameters
    ----------
    key : str
        The official Battlerite API key.
    session : Optional[aiohttp.ClientSession_]
    lang : str, Default['English']
        The language to localise game specific strings in.\n
        Currently available languages are:\n
        `Brazilian, English, French, German, Italian, Japanese, Korean,
        Polish, Romanian, Russian, SChinese, Spanish, Turkish.`
    """
    def __init__(self, key, session: aiohttp.ClientSession=None, lang: str='English'):
        if lang in self.avl_langs:
            self.lang = lang
        else:
            raise Exception('{0} is not an available language.\nAs of now only'
                            'these languages are available:{1}'.format(lang, ', '.join(self.avl_langs)))
        self.session = session or aiohttp.ClientSession()
        self.base_url = "https://api.dc01.gamelockerapp.com/shards/global/"
        self.status_url = "https://api.dc01.gamelockerapp.com/status"
        self.headers = {
            'Authorization': 'Bearer {}'.format(key),
            'Accept': 'application/json'
        }

    async def gen_req(self, url, params=None, session=None):
        sess = session or self.session
        async with sess.get(url, headers=self.headers,
                            params=params) as req:
            try:
                resp = await req.json()
            except (asyncio.TimeoutError, aiohttp.ClientResponseError):
                raise BRRequestException(req, {})

            if 300 > req.status >= 200:
                return resp
            elif req.status == 404:
                raise NotFoundException(req, resp)
            elif req.status > 500:
                raise BRServerException(req, resp)
            else:
                raise BRRequestException(req, resp)

    async def get_status(self):
        """
        Check if the API is up and running

        Returns
        -------
        tuple(createdAt: str, version: str):
        """
        data = await self.gen_req(self.status_url)
        return data['data']['attributes']['releasedAt'], data['data']['attributes']['version']

    async def match_by_id(self, match_id):
        """
        Get a Match by its ID.

        Parameters
        ----------
        match_id : int or str

        Returns
        -------
        :class:`pybattlerite.models.AsyncMatch`
            A match object representing the requested match.
        """
        data = await self.gen_req("{0}matches/{1}".format(self.base_url, match_id))
        return AsyncMatch(data, self.session)

    async def get_matches(self, offset: int=None, limit: int=None, after=None, before=None, playerids: list=None):
        """
        Access the /matches endpoint and grab a list of matches

        .. _datetime.datetime: https://docs.python.org/3.6/library/datetime.html#datetime-objects

        Parameters
        ----------
        offset : Optional[int]
            The nth number of match to start the page from.
        limit : Optional[int]
            Number of matches to return.
        after : Optional[str or datetime.datetime_]
            Filter to return matches after provided time period, if an str is provided it should follow the **iso8601** format.
        before :  Optional[str or datetime.datetime_]
            Filter to return matches before provided time period, if an str is provided it should follow the **iso8601** format.
        playerids : list
            Filter to only return matches with provided players in them by looking for their player IDs.
        Returns
        -------
        :class:`pybattlerite.models.AsyncMatchPaginator`
            A MatchPaginator instance representing a get_matches request
        """
        # Check compatibility 'after' and 'before' with iso8601
        # Also checks if after isn't greater than before

        params = self.prepare_match_params(offset, limit, after, before, playerids)

        data = await self.gen_req("{}matches".format(self.base_url), params=params)
        matches = []
        for match in data['data']:
            matches.append(AsyncMatch(match, self.session, data['included']))
        return AsyncMatchPaginator(matches, data['links'], self)

    async def player_by_id(self, player_id: int):
        """
        Get a player's info by their ID.

        Parameters
        ----------
        player_id : int

        Returns
        -------
        :class:`pybattlerite.models.Player`
            A Player object representing the requested player.
        """
        data = await self.gen_req("{0}players/{1}".format(self.base_url, player_id))
        return Player(data['data'], self.lang)

    async def _players(self, playerids: list=None, steamids: list=None, usernames: list=None, single=False):
        params = self.prepare_players_params(playerids, steamids, usernames)

        data = await self.gen_req("{0}players".format(self.base_url), params=params)
        if len(data['data']) == 0:
            raise EmptyResponseException("No Players with the specified criteria were found.")
        if not single:
            return [Player(player, self.lang) for player in data['data']]
        else:
            return Player(data['data'][0], self.lang)

    async def get_players(self, playerids: list=None, steamids: list=None, usernames: list=None):
        """
        Get multiple players' info at once.

        .. _here: https://developer.valvesoftware.com/wiki/SteamID

        Parameters
        ----------
        playerids : list
            A list of playerids, either a `list` of strs or a `list` of ints.
            Max list length is 6.
        steamids : list
            A list of steamids, a `list` of ints, this accepts only `SteamID64` specification, check here_ for more
            details.
            Max list length is 6
        usernames : list
            A list of usernames, a `list` of strings, case insensitive.
            Max list length is 6

        Returns
        -------
        list
            A list of :class:`pybattlerite.models.Player`
        """
        return await self._players(playerids, steamids, usernames)

    async def player_by_name(self, username):
        """
        Get a player's info by their ingame name.

        Parameters
        ----------
        username : str
            Case insensitive username

        Returns
        -------
        :class:`pybattlerite.models.Player`
            A Player object representing the requested player.
        """
        return await self._players(usernames=[username], single=True)



import requests

from .clientbase import ClientBase
from .models import Player, Match, MatchPaginator
from .errors import BRRequestException
from .errors import NotFoundException
from .errors import BRServerException
from .errors import EmptyResponseException


class Client(ClientBase):
    """
    Top level class for user to interact with the API.

    .. _requests.Session: http://docs.python-requests.org/en/master/api/#request-sessions

    Parameters
    ----------
    key : str
        The official Battlerite API key.
    session : Optional[requests.Session_]
    lang : str, Default['English']
        The language to localise game specific strings in.\n
        Currently available languages are:\n
        `Brazilian, English, French, German, Italian, Japanese, Korean,
        Polish, Romanian, Russian, SChinese, Spanish, Turkish.`
    """
    def __init__(self, key, session: requests.Session=None, lang: str='English'):
        if lang in self.avl_langs:
            self.lang = lang
        else:
            raise Exception('{0} is not an available language.\nAs of now only'
                            'these languages are available:{1}'.format(lang, ', '.join(self.avl_langs)))
        self.session = session or requests.Session()
        self.base_url = "https://api.dc01.gamelockerapp.com/shards/global/"
        self.status_url = "https://api.dc01.gamelockerapp.com/status"
        self.headers = {
            'Authorization': 'Bearer {}'.format(key),
            'Accept': 'application/json'
        }

    def gen_req(self, url, params=None, session=None):
        sess = session or self.session
        with sess.get(url, headers=self.headers,
                      params=params) as req:
            try:
                resp = req.json()
            except (requests.Timeout, requests.ConnectionError):
                raise BRRequestException(req, {})

            if 300 > req.status_code >= 200:
                return resp
            elif req.status_code == 404:
                raise NotFoundException(req, resp)
            elif req.status_code > 500:
                raise BRServerException(req, resp)
            else:
                raise BRRequestException(req, resp)

    def get_status(self):
        """
        Check if the API is up and running

        Returns
        -------
        tuple(createdAt: str, version: str):
        """
        data = self.gen_req(self.status_url)
        return data['data']['attributes']['releasedAt'], data['data']['attributes']['version']

    def match_by_id(self, match_id):
        """
        Get a Match by its ID.

        Parameters
        ----------
        match_id : int or str

        Returns
        -------
        :class:`pybattlerite.models.Match`
            A match object representing the requested match.
        """
        data = self.gen_req("{0}matches/{1}".format(self.base_url, match_id))
        return Match(data, self.session)

    def get_matches(self, offset: int=None, limit: int=None, after=None, before=None, playerids: list=None):
        """
        Access the /matches endpoint and grab a list of matches

        .. _datetime.datetime: https://docs.python.org/3.6/library/datetime.html#datetime-objects

        Parameters
        ----------
        offset : Optional[int]
            The nth number of match to start the page from.
        limit : Optional[int]
            Number of matches to return.
        after : Optional[str or datetime.datetime_]
            Filter to return matches after provided time period, if an str is provided it should follow the **iso8601** format.
        before :  Optional[str or datetime.datetime_]
            Filter to return matches before provided time period, if an str is provided it should follow the **iso8601** format.
        playerids : list
            Filter to only return matches with provided players in them by looking for their player IDs.
        Returns
        -------
        :class:`pybattlerite.models.MatchPaginator`
            A MatchPaginator instance representing a get_matches request
        """

        # Check compatibility 'after' and 'before' with iso8601
        # Also checks if after isn't greater than before
        params = self.prepare_match_params(offset, limit, after, before, playerids)

        data = self.gen_req("{}matches".format(self.base_url), params=params)
        matches = []
        for match in data['data']:
            matches.append(Match(match, self.session, data['included']))
        return MatchPaginator(matches, data['links'], self)

    def player_by_id(self, player_id: int):
        """
        Get a player's info by their ID.

        Parameters
        ----------
        player_id : int

        Returns
        -------
        :class:`pybattlerite.models.Player`
            A Player object representing the requested player.
        """
        data = self.gen_req("{0}players/{1}".format(self.base_url, player_id))
        return Player(data['data'], self.lang)

    def _players(self, playerids: list=None, steamids: list=None, usernames: list=None, single=False):
        params = self.prepare_players_params(playerids, steamids, usernames)
        data = self.gen_req("{0}players".format(self.base_url), params=params)
        if len(data['data']) == 0:
            raise EmptyResponseException("No Players with the specified criteria were found.")
        if not single:
            return [Player(player, self.lang) for player in data['data']]
        else:
            return Player(data['data'][0], self.lang)

    def get_players(self, playerids: list=None, steamids: list=None, usernames: list=None):
        """
        Get multiple players' info at once.

        .. _here: https://developer.valvesoftware.com/wiki/SteamID

        Parameters
        ----------
        playerids : list
            A list of playerids, either a `list` of strs or a `list` of ints.
            Max list length is 6.
        steamids : list
            A list of steamids, a `list` of ints, this accepts only `SteamID64` specification, check here_ for more
            details.
            Max list length is 6
        usernames : list
            A list of usernames, a `list` of strings, case insensitive.
            Max list length is 6

        Returns
        -------
        list
            A list of :class:`pybattlerite.models.Player`
        """
        return self._players(playerids, steamids, usernames)

    def player_by_name(self, username):
        """
        Get a player's info by their ingame name.

        Parameters
        ----------
        username : str
            Case insensitive username

        Returns
        -------
        :class:`pybattlerite.models.Player`
            A Player object representing the requested player.
        """
        return self._players(usernames=[username], single=True)



from setuptools import setup

setup(
    name="pybattlerite",
    packages=['pybattlerite'],
    version="0.5.16",
    description="A wrapper for the madglory Battlerite API, complete with synchronous and asynchronous client implementations.",
    author="xKynn",
    author_email="xkynn@github.com",
    url="https://github.com/xKynn/PyBattlerite",
    download_url="https://github.com/xKynn/PyBattlerite/archive/0.5.16.tar.gz",
    keywords=['battlerite', 'asyncbattlerite', 'async-battlerite', 'async_battlerite'],
    classifiers=[],
    install_requires=[
        "aiohttp",
        "requests"
    ],
    python_requires='>=3.5',
    package_data={
        '': ['data/*.json', 'data/localization/*.ini']
    }
)



class BRRequestException(Exception):
    """
    General purpose exception, base class for other request related exceptions.

    .. _aiohttp.ClientResponse : https://aiohttp.readthedocs.io/en/stable/client_reference.html#aiohttp.ClientResponse

    Parameters
    ----------
    response : aiohttp.ClientResponse_
    data : dict
        The json response from the API
    """
    def __init__(self, response, data):
        self.reason = response.reason
        self.status = response.status
        self.error = data.get("errors")
        if self.error is not None:
            self.error = self.error[0]['title']
        super().__init__("{0.status}: {0.reason} - {0.error}".format(self))


class NotFoundException(BRRequestException):
    """
    For the 404s
    """
    pass


class BRServerException(BRRequestException):
    """
    Exception that signifies that the server failed to respond with valid data.
    """
    pass


class BRFilterException(Exception):
    """
    Raised when an invalid filter value is supplied.
    """
    def __init__(self, error):
        super().__init__(error)


class BRPaginationError(Exception):
    """
    Raised when :meth:`asyncbattlerite.models.MatchPaginator.next` or :meth:`asyncbattlerite.models.MatchPaginator.prev`
    are called when the paginator is on the last or first page respectively.
    """
    def __init__(self, error):
        super().__init__(error)


class EmptyResponseException(Exception):
    """
    Raised when any request is 200 OK, but the data is empty.
    """
    def __init__(self, error):
        super().__init__(error)